'''
(C) Copyright 2019 Hewlett Packard Enterprise Development LP

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
'''

# Imports from python standard library
from base64 import b64decode
from datetime import datetime, timezone
from urllib.request import urlopen

import asyncio
import pwd
import subprocess

# Imports to work with JupyterHub
from jupyterhub.auth import Authenticator
from jupyterhub.utils import maybe_future
from jupyterhub.handlers.base import BaseHandler
from jupyterhub.handlers.login import LoginHandler, LogoutHandler
from tornado import gen, web
from traitlets import Unicode, Bool
from jinja2 import Template

# Imports for me
from lxml import etree
import pytz
from signxml import XMLVerifier

class SAMLAuthenticator(Authenticator):
    metadata_filepath = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A filepath to the location of the SAML IdP metadata. This is the most preferable
        option for presenting an IdP's metadata to the authenticator.
        '''
    )
    metadata_content = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A fully-inlined version of the SAML IdP metadata. Mostly provided for testing,
        but if you want to use this for a "production-type" system, I'm not going to
        judge. This is preferred above getting metadata from a web-request, but not
        preferred above getting the metadata from a file.
        '''
    )
    metadata_url = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A URL where the SAML Authenticator can find metadata for the SAML IdP. This is
        the least preferable method of providing the SAML IdP metadata to the
        authenticator, as it is both slow and vulnerable to Man in the Middle attacks,
        including DNS poisoning.
        '''
    )
    xpath_username_location = Unicode(
        default_value='//saml:NameID/text()',
        allow_none=True,
        config=True,
        help='''
        This is an XPath that specifies where the user's name or id is located in the
        SAML Assertion. This is partly for testing purposes, but there are cases where
        an administrator may want a user to be identified by their email address instead
        of an LDAP DN or another string that comes in the NameID field. The namespace
        bindings when executing the XPath will be as follows:

        {
            'ds'   : 'http://www.w3.org/2000/09/xmldsig#',
            'md'   : 'urn:oasis:names:tc:SAML:2.0:metadata',
            'saml' : 'urn:oasis:names:tc:SAML:2.0:assertion',
            'samlp': 'urn:oasis:names:tc:SAML:2.0:protocol'
        }
        '''
    )
    login_post_field = Unicode(
        default_value='SAMLResponse',
        allow_none=False,
        config=True,
        help='''
        This value specifies what field in the SAML Post request contains the Base-64
        encoded SAML Response.
        '''
    )
    audience = Unicode(
        default_value=None,
        allow_none=True,
        config=True,
        help='''
        The SAML Audience must be configured in the SAML IdP. This value ensures that a
        SAML assertion cannot be used by a malicious service to authenticate to a naive
        service. If this value is not set in the configuration file or if the string
        provided is a "false-y" value in python, this will not be checked.
        '''
    )
    recipient = Unicode(
        default_value=None,
        allow_none=True,
        config=True,
        help='''
        The SAML Recipient must be configured in the SAML IdP. This value ensures that a
        SAML assertion cannot be used by a malicious service to authenticate to a naive
        service. If this value is not set in the configuration file or if the string
        provided is a "false-y" value in python, this will not be checked.
        '''
    )
    time_format_string = Unicode(
        default_value='%Y-%m-%dT%H:%M:%SZ',
        allow_none=False,
        config=True,
        help='''
        A time format string that complies with python's strftime()/strptime() behavior.
        For more information on this format, please read the information at the
        following link:

        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior

        '''
    )
    idp_timezone = Unicode(
        default_value='UTC',
        allow_none=True,
        config=True,
        help='''
        A timezone-specific string that uniquely identifies a timezone using pytz's
        timezone constructor. To view a list of options, import the package and
        inspect the `pytz.all_timezones` list. It is quite long. For more information
        on pytz, please read peruse the pip package:

        https://pypi.org/project/pytz/

        '''
    )
    shutdown_on_logout = Bool(
        default_value=False,
        allow_none=False,
        config=True,
        help='''
        If you would like to shutdown user servers on logout, you can enable this
        behavior with:

        c.SAMLAuthenticator.shutdown_on_logout = True

        Be careful with this setting because logging out one browser does not mean
        the user is no longer actively using their server from another machine.

        It is a little odd to have this property on the Authenticator object, but
        (for internal-detail-reasons) since we need to hand-craft the LogoutHandler
        class, this should be on the Authenticator.
        '''
    )
    slo_forwad_on_logout = Bool(
        default_value=True,
        allow_none=False,
        config=True,
        help='''
        [DEPRECATED] Please use `slo_forward_on_logout`.
        This attribute will be removed in the next version.
        See https://github.com/bluedatainc/jupyterhub-samlauthenticator/releases/tag/samlauthenticator-0.0.7
        for more information.
        '''
    )
    slo_forward_on_logout = Bool(
        default_value=True,
        allow_none=False,
        config=True,
        help='''
        To prevent forwarding users to the SLO URI on logout,
        set this parameter to False like so:

        c.SAMLAuthenticator.slo_forward_on_logout = False
        '''
    )
    entity_id = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        The entity id for this specific JupyterHub instance. If
        populated, this will be included in the SP metadata as
        the entity id. If this is not populated, the entity will
        populate as the protocol, host, and port of the request
        to get the SAML Metadata.

        Note that if the JupyterHub server will be behind a
        proxy, this should be populated as the protocol, host,
        and port where the server can be reached. For example,
        if the JupyterHub server should be reached at
        10.0.31.2:8000, this should be populated as
        'https://10.0.31.2:8000'
        '''
    )
    nameid_format = Unicode(
        default_value='urn:oasis:names:tc:SAML:2.0:nameid-format:transient',
        allow_none=True,
        config=True,
        help='''
        The nameId format to set in the Jupyter SAML Metadata.
        Defaults to transient nameid-format, but other values such as
        urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress or
        urn:oasis:names:tc:SAML:2.0:nameid-format:persistent
        are available. See section 8.3 of the spec
        http://docs.oasis-open.org/security/saml/v2.0/saml-core-2.0-os.pdf
        for more details.
        '''
    )
    acs_endpoint_url = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        The access consumer endpoint url for this specific
        JupyterHub instance. If populated, this will be
        included in the SP metadata as the acs endpoint
        location. If populated, this field MUST tell the
        SAML IdP to post to the ip address and port the
        JupyterHub is running on concatenated to
        "/hub/login". For example, if the server were
        running on 10.0.31.2:8000, this value should be
        'https://10.0.31.2:8000/hub/login'. It is necessary
        to populate this field if the ACS Endpoint is
        significantly different from the entity id.
        If this is not populated, the entity location
        will populate as the entity id concatenated
        to '/hub/login'.
        '''
    )
    organization_name = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A short-form organization name. Will be populated into the
        SP metadata.
        '''
    )
    organization_display_name = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A long-form organization name. Will be populated into the
        SP metadata.
        '''
    )
    organization_url = Unicode(
        default_value='',
        allow_none=True,
        config=True,
        help='''
        A URL that uniquely identifies the organization.
        '''
    )
    create_system_users = Bool(
        default_value=True,
        allow_none=False,
        config=True,
        help='''
        When True, SAMLAuthenticator will create system users
        on user authentication if they don't exist already.
        Default value is True.
        '''
    )
    create_system_user_binary = Unicode(
        default_value='useradd',
        allow_none=True,
        config=True,
        help='''
        When SAMLAuthenticator creates a system user (also called "just in time user provisioning")
        it calls the binary specified in this property in a subprocess to perform the user creation.
        Default value is 'useradd'.
        This can be set to any binary in the host machine's PATH or a full path to an alternate
        binary not in the host's path. This binary MUST accpet calls of the form
        "${binary_name} ${user_name}" and exit with a status of zero on valid user addition or
        a non-zero status in the failure case.
        '''
    )
    xpath_role_location = Unicode(
        default_value=None,
        allow_none=True,
        config=True,
        help='''
        This is an XPath that specifies where the user's roles are located in
        the SAML  Assertion. This is to restrict users with certain roles
        granted by the administrator to have access to jupyterhub.
        '''
    )
    allowed_roles = Unicode(
        default_value=None,
        allow_none=True,
        config=True,
        help='''
        Comma-separated list of roles. SAMLAuthenticator will restrict access to
        jupyterhub to these roles if specified.
        '''
    )
    _const_warn_explain       = 'Because no user would be allowed to log in via roles, role check disabled.'
    _const_warn_no_role_xpath = 'Allowed roles set while role location XPath is not set.'
    _const_warn_no_roles      = 'Allowed roles not set while role location XPath is set.'

    def _get_metadata_from_file(self):
        with open(self.metadata_filepath, 'r') as saml_metadata:
            return saml_metadata.read()

    def _get_metadata_from_config(self):
        return self.metadata_content

    def _get_metadata_from_url(self):
        with urlopen(self.metadata_url) as remote_metadata:
            return remote_metadata.read()

    def _get_preferred_metadata_from_source(self):
        if self.metadata_filepath:
            return self._get_metadata_from_file()

        if self.metadata_content:
            return self._get_metadata_from_config()

        if self.metadata_url:
            return self._get_metadata_from_url()

        return None

    def _log_exception_error(self, exception):
        self.log.warning('Exception: %s', str(exception))

    def _get_saml_doc_etree(self, data):
        saml_response = data.get(self.login_post_field, None)

        if not saml_response:
            # Failed to get the SAML Response from the posted data
            self.log.warning('Could not get SAML Response from post data')
            self.log.warning('Expected SAML response in field %s', self.login_post_field)
            self.log.warning('Posted login data %s', str(data))
            return None

        decoded_saml_doc = None

        try:
            decoded_saml_doc = b64decode(saml_response)
        except Exception as e:
            # There was a problem base64 decoding the xml document from the posted data
            self.log.warning('Got exception when attempting to decode SAML response')
            self.log.warning('Saml Response: %s', saml_response)
            self._log_exception_error(e)
            return None

        try:
            return etree.fromstring(decoded_saml_doc)
        except Exception as e:
            self.log.warning('Got exception when attempting to hydrate response to etree')
            self.log.warning('Saml Response: %s', decoded_saml_doc)
            self._log_exception_error(e)
            return None

    def _get_saml_metadata_etree(self):
        try:
            saml_metadata = self._get_preferred_metadata_from_source()
        except Exception as e:
            # There was a problem getting the SAML metadata
            self.log.warning('Got exception when attempting to read SAML metadata')
            self.log.warning('Ensure that EXACTLY ONE of metadata_filepath, ' +
                             'metadata_content, and metadata_url is populated')
            self._log_exception_error(e)
            return None

        if not saml_metadata:
            # There was a problem getting the SAML metadata
            self.log.warning('Got exception when attempting to read SAML metadata')
            self.log.warning('Ensure that EXACTLY ONE of metadata_filepath, ' +
                             'metadata_content, and metadata_url is populated')
            self.log.warning('SAML metadata was empty')
            return None

        metadata_etree = None

        try:
            metadata_etree = etree.fromstring(saml_metadata)
        except Exception as e:
            # Failed to parse SAML Metadata
            self.log.warning('Got exception when attempting to parse SAML metadata')
            self._log_exception_error(e)

        return metadata_etree

    def _verify_saml_signature(self, saml_metadata, decoded_saml_doc):
        xpath_with_namespaces = self._make_xpath_builder()
        find_cert = xpath_with_namespaces('//ds:KeyInfo/ds:X509Data/ds:X509Certificate/text()')
        cert_value = None

        try:
            cert_value = find_cert(saml_metadata)[0]
        except Exception as e:
            self.log.warning('Could not get cert value from saml metadata')
            self._log_exception_error(e)
            return None

        signed_xml = None
        try:
            signed_xml = XMLVerifier().verify(decoded_saml_doc, x509_cert=cert_value).signed_xml
        except Exception as e:
            self.log.warning('Failed to verify signature on SAML Response')
            self._log_exception_error(e)

        return signed_xml

    def _make_xpath_builder(self):
        namespaces = {
            'ds'   : 'http://www.w3.org/2000/09/xmldsig#',
            'md'   : 'urn:oasis:names:tc:SAML:2.0:metadata',
            'saml' : 'urn:oasis:names:tc:SAML:2.0:assertion',
            'samlp': 'urn:oasis:names:tc:SAML:2.0:protocol'
        }

        def xpath_with_namespaces(xpath_str):
            return etree.XPath(xpath_str, namespaces=namespaces)

        return xpath_with_namespaces

    def _verify_saml_response_against_metadata(self, saml_metadata, signed_xml):
        xpath_with_namespaces = self._make_xpath_builder()

        find_entity_id = xpath_with_namespaces('//saml:Issuer/text()')
        find_metadata_entity_id = xpath_with_namespaces('//md:EntityDescriptor/@entityID')

        saml_metadata_entity_id_list = find_metadata_entity_id(saml_metadata)
        saml_resp_entity_id_list = find_entity_id(signed_xml)

        if saml_resp_entity_id_list and saml_metadata_entity_id_list:
            if saml_metadata_entity_id_list[0] != saml_resp_entity_id_list[0]:
                self.log.warning('Metadata entity id did not match the response entity id')
                self.log.warning('Metadata entity id: %s', saml_metadata_entity_id_list[0])
                self.log.warning('Response entity id: %s', saml_resp_entity_id_list[0])
                return False
        else:
            self.log.warning('The entity ID needs to be set in both the metadata and the SAML Response')
            if not saml_resp_entity_id_list:
                self.log.warning('The entity ID was not set in the SAML Response')
            if not saml_metadata_entity_id_list:
                self.log.warning('The entity ID was not set in the SAML metadata')
            return False

        return True

    def _verify_saml_response_against_configured_fields(self, signed_xml):
        xpath_with_namespaces = self._make_xpath_builder()

        if self.audience:
            find_audience = xpath_with_namespaces('//saml:Audience/text()')
            saml_resp_audience_list = find_audience(signed_xml)
            if saml_resp_audience_list:
                if saml_resp_audience_list[0] != self.audience:
                    self.log.warning('Configured audience did not match the response audience')
                    self.log.warning('Configured audience: %s', self.audience)
                    self.log.warning('Response audience: %s', saml_resp_audience_list[0])
                    return False
            else:
                self.log.warning('SAML Audience was set in authenticator config file, but not in SAML Response')
                return False

        if self.recipient:
            find_recipient = xpath_with_namespaces('//saml:SubjectConfirmationData/@Recipient')
            recipient_list = find_recipient(signed_xml)
            if recipient_list:
                if self.recipient != recipient_list[0]:
                    self.log.warning('Configured recipient did not match the response recipient')
                    self.log.warning('Configured recipient: %s', self.recipient)
                    self.log.warning('Response recipient: %s', recipient_list[0])
                    return False
            else:
                self.log.warning('Could not find recipient in SAML response')
                return False

        return True

    def _is_date_aware(self, created_datetime):
        return created_datetime.tzinfo is not None and \
            created_datetime.tzinfo.utcoffset(created_datetime) is not None

    def _verify_physical_constraints(self, signed_xml):
        xpath_with_namespaces = self._make_xpath_builder()

        find_not_before = xpath_with_namespaces('//saml:Conditions/@NotBefore')
        find_not_on_or_after = xpath_with_namespaces('//saml:Conditions/@NotOnOrAfter')

        not_before_list = find_not_before(signed_xml)
        not_on_or_after_list = find_not_on_or_after(signed_xml)

        if not_before_list and not_on_or_after_list:

            not_before_datetime = datetime.strptime(not_before_list[0], self.time_format_string)
            not_on_or_after_datetime = datetime.strptime(not_on_or_after_list[0], self.time_format_string)

            timezone_obj = None

            if not self._is_date_aware(not_before_datetime):
                timezone_obj = pytz.timezone(self.idp_timezone)
                not_before_datetime = timezone_obj.localize(not_before_datetime)

            if not self._is_date_aware(not_on_or_after_datetime):
                if not timezone_obj:
                    timezone_obj = pytz.timezone(self.idp_timezone)
                not_on_or_after_datetime = timezone_obj.localize(not_on_or_after_datetime)

            now = datetime.now(timezone.utc)

            if now < not_before_datetime or now >= not_on_or_after_datetime:
                self.log.warning('Bad timing condition')
                if now < not_before_datetime:
                    self.log.warning('Sent SAML Response before it was permitted')
                if now >= not_on_or_after_datetime:
                    self.log.warning('Sent SAML Response after it was permitted')
                return False
        else:
            self.log.warning('SAML assertion did not contain proper conditions')
            if not not_before_list:
                self.log.warning('SAML assertion must have NotBefore annotation in Conditions')
            if not not_on_or_after_list:
                self.log.warning('SAML assertion must have NotOnOrAfter annotation in Conditions')
            return False

        return True

    def _verify_saml_response_fields(self, saml_metadata, signed_xml):
        if not self._verify_saml_response_against_metadata(saml_metadata, signed_xml):
            self.log.warning('The SAML Assertion did not match the provided metadata')
            return False

        if not self._verify_saml_response_against_configured_fields(signed_xml):
            self.log.warning('The SAML Assertion did not match the configured values')
            return False

        if not self._verify_physical_constraints(signed_xml):
            self.log.warning('The SAML Assertion did not match the physical constraints')
            return False

        self.log.info('The SAML Assertion matched the configured values')
        return True

    def _test_valid_saml_response(self, saml_metadata, saml_doc):
        signed_xml = self._verify_saml_signature(saml_metadata, saml_doc)

        if signed_xml is None or len(signed_xml) == 0:
            self.log.warning('Failed to verify signature on SAML Response')
            return False, None

        return self._verify_saml_response_fields(saml_metadata, signed_xml), signed_xml

    def _get_username_from_saml_etree(self, signed_xml):
        xpath_with_namespaces = self._make_xpath_builder()

        xpath_fun = xpath_with_namespaces(self.xpath_username_location)
        xpath_result = xpath_fun(signed_xml)

        if isinstance(xpath_result, etree._ElementUnicodeResult):
            return xpath_result
        if type(xpath_result) is list and len(xpath_result) > 0:
            return xpath_result[0]

        self.log.warning('Could not find name from name XPath')
        return None

    def _get_roles_from_saml_etree(self, signed_xml):
        if self.xpath_role_location:
            xpath_with_namespaces = self._make_xpath_builder()
            xpath_fun = xpath_with_namespaces(self.xpath_role_location)
            xpath_result = xpath_fun(signed_xml)

            if xpath_result:
                return xpath_result

            self.log.warning('Could not find role from role XPath')

        return []

    def _get_username_from_saml_doc(self, signed_xml, decoded_saml_doc):
        user_name = self._get_username_from_saml_etree(signed_xml)
        if user_name:
            return user_name

        self.log.info('Did not get user name from signed SAML Response')

        return self._get_username_from_saml_etree(decoded_saml_doc)

    def _get_roles_from_saml_doc(self, signed_xml, decoded_saml_doc):
        user_roles = self._get_roles_from_saml_etree(signed_xml)
        if user_roles:
            return user_roles

        self.log.info('Did not get user roles from signed SAML Response')

        return self._get_roles_from_saml_etree(decoded_saml_doc)

    def _optional_user_add(self, username):
        try:
            pwd.getpwnam(username)
            # Found the user, we don't need to create them
            return True
        except KeyError:
            # Return the `not` here because a 0 return indicates success and I want to
            # say something like "if adding the user is successful, return username"
            return not subprocess.call([self.create_system_user_binary, username])

    def _check_username_and_add_user(self, username):
        if self.validate_username(username) and \
                self.check_blacklist(username) and \
                self.check_whitelist(username):
            if self.create_system_users:
                if self._optional_user_add(username):
                    # Successfully added user
                    return username
                else:
                    # Failed to add user
                    self.log.error('Failed to add user by calling add user')
                    return None

            # Didn't try to add user
            return username

        # Failed to validate username or failed list check
        self.log.error('Failed to validate username or failed list check')
        return None

    def _check_role(self, user_roles):
        allowed_roles = [x.strip() for x in self.allowed_roles.split(',')]

        return any(elem in allowed_roles for elem in user_roles)

    def _valid_roles_in_assertion(self, signed_xml, saml_doc_etree):
        user_roles = self._get_roles_from_saml_doc(signed_xml, saml_doc_etree)

        user_roles_result = self._check_role(user_roles)
        if not user_roles_result:
            self.log.error('User role not authorized')
        return user_roles_result

    def _valid_config_and_roles(self, signed_xml, saml_doc_etree):
        if self.allowed_roles and self.xpath_role_location:
            return self._valid_roles_in_assertion(signed_xml, saml_doc_etree)

        if (not self.allowed_roles) and self.xpath_role_location:
            self.log.warning(self._const_warn_no_roles)
            self.log.warning(self._const_warn_explain)

        if self.allowed_roles and (not self.xpath_role_location):
            self.log.warning(self._const_warn_no_role_xpath)
            self.log.warning(self._const_warn_explain)

        # This technically skips the "neither set" case, but since that's expected-ish, I think we can let
        # that slide.
        return True

    def _authenticate(self, handler, data):
        saml_doc_etree = self._get_saml_doc_etree(data)

        if saml_doc_etree is None or len(saml_doc_etree) == 0:
            self.log.error('Error getting decoded SAML Response')
            return None

        saml_metadata_etree = self._get_saml_metadata_etree()

        if saml_metadata_etree is None or len(saml_metadata_etree) == 0:
            self.log.error('Error getting SAML Metadata')
            return None

        valid_saml_response, signed_xml = self._test_valid_saml_response(saml_metadata_etree, saml_doc_etree)

        if valid_saml_response:
            self.log.debug('Authenticated user using SAML')
            username = self._get_username_from_saml_doc(signed_xml, saml_doc_etree)
            username = self.normalize_username(username)

            if self._valid_config_and_roles(signed_xml, saml_doc_etree):
                self.log.debug('Optionally create and return user: ' + username)
                return self._check_username_and_add_user(username)

            self.log.error('Assertion did not have appropriate roles')
            return None

        self.log.error('Error validating SAML response')
        return None

    @gen.coroutine
    def authenticate(self, handler, data):
        return self._authenticate(handler, data)

    def _get_redirect_from_metadata_and_redirect(authenticator_self, element_name, handler_self):
        saml_metadata_etree = authenticator_self._get_saml_metadata_etree()

        handler_self.log.debug('Got metadata etree')

        if saml_metadata_etree is None or len(saml_metadata_etree) == 0:
            handler_self.log.error('Error getting SAML Metadata')
            raise web.HTTPError(500)

        handler_self.log.debug('Got valid metadata etree')

        xpath_with_namespaces = authenticator_self._make_xpath_builder()

        binding = 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
        final_xpath = '//' + element_name + '[@Binding=\'' + binding + '\']/@Location'
        handler_self.log.debug('Final xpath is: ' + final_xpath)

        redirect_link_getter = xpath_with_namespaces(final_xpath)

        # Here permanent MUST BE False - otherwise the /hub/logout GET will not be fired
        # by the user's browser.
        handler_self.redirect(redirect_link_getter(saml_metadata_etree)[0], permanent=False)

    def _make_org_metadata(self):
        if self.organization_name or \
                self.organization_display_name or \
                self.organization_url:
            org_name_elem = org_disp_name_elem = org_url_elem = ''
            organization_name_element = '''<OrganizationName>{{ name }}</OrganizationName>'''
            organization_display_name_element = '''<OrganizationDisplayName>{{ displayName }}</OrganizationDisplayName>'''
            organization_url_element = '''<OrganizationURL>{{ url }}</OrganizationURL>'''
            organization_metadata = '''
    <Organization>
        {{ organizationName }}
        {{ organizationDisplayName }}
        {{ organizationUrl }}
    </Organization>
    '''

            if self.organization_name:
                org_name_template = Template(organization_name_element)
                org_name_elem = org_name_template.render(name=self.organization_name)

            if self.organization_display_name:
                org_disp_name_template = Template(organization_display_name_element)
                org_disp_name_elem = org_disp_name_template.render(displayName=self.organization_display_name)

            if self.organization_url:
                org_url_template = Template(organization_url_element)
                org_url_elem = org_url_template.render(url=self.organization_url)

            org_metadata_template = Template(organization_metadata)
            return org_metadata_template.render(organizationName=org_name_elem,
                                                organizationDisplayName=org_disp_name_elem,
                                                organizationUrl=org_url_elem)

        return ''

    def _make_sp_metadata(authenticator_self, meta_handler_self):
        metadata_text = '''<?xml version="1.0"?>
<EntityDescriptor
        entityID="{{ entityId }}"
        xmlns="urn:oasis:names:tc:SAML:2.0:metadata"
        xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
    <SPSSODescriptor
            AuthnRequestsSigned="false"
            protocolSupportEnumeration="urn:oasis:names:tc:SAML:2.0:protocol">
        <NameIDFormat>
            {{ nameIdFormat }}
        </NameIDFormat>
        <AssertionConsumerService
                Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                Location="{{ entityLocation }}"/>
    </SPSSODescriptor>
    {{ organizationMetadata }}
</EntityDescriptor>
'''

        entity_id = authenticator_self.entity_id if authenticator_self.entity_id else \
                meta_handler_self.request.protocol + '://' + meta_handler_self.request.host

        acs_endpoint_url = authenticator_self.acs_endpoint_url if authenticator_self.acs_endpoint_url else \
                entity_id + '/hub/login'

        org_metadata_elem = authenticator_self._make_org_metadata()

        xml_template = Template(metadata_text)
        return xml_template.render(entityId=entity_id,
                                   nameIdFormat=authenticator_self.nameid_format,
                                   entityLocation=acs_endpoint_url,
                                   organizationMetadata=org_metadata_elem)

    def get_handlers(authenticator_self, app):

        class SAMLLoginHandler(LoginHandler):

            async def get(login_handler_self):
                login_handler_self.log.info('Starting SP-initiated SAML Login')
                authenticator_self._get_redirect_from_metadata_and_redirect('md:SingleSignOnService',
                                                                            login_handler_self)

        class SAMLLogoutHandler(LogoutHandler):
            # TODO: When the time is right to force users onto JupyterHub 1.0.0,
            # refactor this.
            async def _shutdown_servers(self, user):
                active_servers = [
                    name
                    for (name, spawner) in user.spawners.items()
                    if spawner.active and not spawner.pending
                ]
                if active_servers:
                    self.log.debug("Shutting down %s's servers", user.name)
                    futures = []
                    for server_name in active_servers:
                        futures.append(maybe_future(self.stop_single_user(user, server_name)))
                    await asyncio.gather(*futures)

            def _backend_logout_cleanup(self, name):
                self.log.info("User logged out: %s", name)
                self.clear_login_cookie()
                self.statsd.incr('logout')

            async def _shutdown_servers_and_backend_cleanup(self):
                user = self.current_user
                if user:
                    await self._shutdown_servers(user)

            async def get(logout_handler_self):
                if authenticator_self.shutdown_on_logout:
                    logout_handler_self.log.debug('Shutting down servers during SAML Logout')
                    await logout_handler_self._shutdown_servers_and_backend_cleanup()

                if logout_handler_self.current_user:
                    logout_handler_self._backend_logout_cleanup(logout_handler_self.current_user.name)

                # This is a little janky, but there was a misspelling in a prior version
                # where someone could have set the wrong flag because of the documentation.
                # We will honor the misspelling until we rev the version, and then we will
                # break backward compatibility.
                forward_on_logout = True if authenticator_self.slo_forward_on_logout else False
                forwad_on_logout = True if authenticator_self.slo_forwad_on_logout else False
                if forward_on_logout or forwad_on_logout:
                    authenticator_self._get_redirect_from_metadata_and_redirect('md:SingleLogoutService',
                                                                                logout_handler_self)
                else:
                    html = logout_handler_self.render_template('logout.html', sync=True)
                    logout_handler_self.finish(html)

        class SAMLMetaHandler(BaseHandler):

            async def get(meta_handler_self):
                xml_content = authenticator_self._make_sp_metadata(meta_handler_self)
                meta_handler_self.set_header('Content-Type', 'text/xml')
                meta_handler_self.write(xml_content)


        return [('/login', SAMLLoginHandler),
                ('/hub/login', SAMLLoginHandler),
                ('/logout', SAMLLogoutHandler),
                ('/hub/logout', SAMLLogoutHandler),
                ('/metadata', SAMLMetaHandler),
                ('/hub/metadata', SAMLMetaHandler)]



'''
(C) Copyright 2019 Hewlett Packard Enterprise Development LP

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the 'Software'),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
'''
from datetime import datetime, timezone
from unittest.mock import patch, MagicMock

import pytest
import unittest

from samlauthenticator import SAMLAuthenticator

from jupyterhub.handlers.base import BaseHandler
from lxml import etree
from signxml import XMLVerifier
from tornado.web import HTTPError

from . import test_constants


class TestMetadataRetrieval(unittest.TestCase):
    # TODO: move metadata xml inside this object
    def _test_high_level_metadata_retrieval_functions(self, authenticator):
        assert authenticator._get_preferred_metadata_from_source() == test_constants.sample_metadata_xml

        metadata_etree = authenticator._get_saml_metadata_etree()
        local_etree = etree.fromstring(test_constants.sample_metadata_xml)

        assert etree.tostring(metadata_etree) == etree.tostring(local_etree)

    def _test_readable_mock(self, authenticator, mock_obj):
        # Make sure that we called this from inside a context handler
        # so that if it exits unexpectedly, we don't have to do manual cleanup
        mock_obj().__enter__.assert_called_once()
        mock_obj().__exit__.assert_called_once()

        self._test_high_level_metadata_retrieval_functions(authenticator)

    @patch('samlauthenticator.samlauthenticator.open')
    def test_file_read(self, mock_fileopen):
        entered_obj = MagicMock()
        entered_obj.read.return_value = test_constants.sample_metadata_xml
        mock_fileopen().__enter__.return_value = entered_obj

        a = SAMLAuthenticator()
        a.metadata_url = 'bad_data'
        a.metadata_content = 'bad_data'
        a.metadata_filepath = '/completely/legitimate/filepath'

        assert a._get_metadata_from_file() == test_constants.sample_metadata_xml
        # Check that we have, at least once, called open with the provided filepath
        # TODO: Figure out how to do this so we can use 'assert_called_once_with'
        mock_fileopen.assert_any_call(a.metadata_filepath, 'r')
        # Check that we're reading the file
        entered_obj.read.assert_called_once()

        self._test_readable_mock(a, mock_fileopen)

    def test_metadata_field(self):
        a = SAMLAuthenticator()
        a.metadata_url = 'bad_data'
        a.metadata_content = test_constants.sample_metadata_xml

        assert a._get_metadata_from_config() == test_constants.sample_metadata_xml
        self._test_high_level_metadata_retrieval_functions(a)

    @patch('samlauthenticator.samlauthenticator.urlopen')
    def test_metadata_url(self, mock_urlopen):
        entered_obj = MagicMock()
        entered_obj.read.return_value = test_constants.sample_metadata_xml
        mock_urlopen().__enter__.return_value = entered_obj

        a = SAMLAuthenticator()
        a.metadata_url = 'http://foo'

        # Check that we're getting the right value
        assert a._get_metadata_from_url() == test_constants.sample_metadata_xml
        # Check that we have, at least once, called open with the provided url
        # TODO: Figure out how to do this so we can use 'assert_called_once_with'
        mock_urlopen.assert_any_call(a.metadata_url)
        # Check that we're reading the file
        entered_obj.read.assert_called_once()

        self._test_readable_mock(a, mock_urlopen)

    @patch('samlauthenticator.samlauthenticator.open')
    def test_file_fail(self, mock_fileopen):
        entered_obj = MagicMock()
        entered_obj.read.side_effect = IOError('Fake IO Error')
        mock_fileopen().__enter__.return_value = entered_obj

        a = SAMLAuthenticator()
        a.metadata_url = 'bad_data'
        a.metadata_content = 'bad_data'
        a.metadata_filepath = '/completely/illegitimate/filepath'

        with pytest.raises(IOError):
            a._get_metadata_from_file()

        with pytest.raises(IOError):
            a._get_preferred_metadata_from_source()

        assert a._get_saml_metadata_etree() is None

    @patch('samlauthenticator.samlauthenticator.urlopen')
    def test_urlopen_fail(self, mock_urlopen):
        entered_obj = MagicMock()
        entered_obj.read.side_effect = IOError('Fake IO Error')
        mock_urlopen().__enter__.return_value = entered_obj

        a = SAMLAuthenticator()
        a.metadata_url = 'http://foo'

        with pytest.raises(IOError):
            a._get_metadata_from_url()

        with pytest.raises(IOError):
            a._get_preferred_metadata_from_source()

        assert a._get_saml_metadata_etree() is None

    def test_no_metadata_configured(self):
        a = SAMLAuthenticator()
        assert a._get_preferred_metadata_from_source() is None

        assert a._get_saml_metadata_etree() is None

    def test_malformed_metadata(self):
        a = SAMLAuthenticator()
        bad_xml = 'not an xml document'
        a.metadata_content = bad_xml

        assert a._get_metadata_from_config() == bad_xml
        assert a._get_preferred_metadata_from_source() == bad_xml
        assert a._get_saml_metadata_etree() is None


class TestSAMLDocRetrieval(unittest.TestCase):
    # TODO: move SAMLResponse inside this object
    def test_get_saml_doc_etree(self):
        # We expect the SAML Response to be coming in base 64 encoded
        a = SAMLAuthenticator()
        fake_data = {a.login_post_field: test_constants.b64encoded_response_xml}

        faked_etree = a._get_saml_doc_etree(fake_data)
        real_etree = etree.fromstring(test_constants.sample_response_xml)

        assert etree.tostring(faked_etree) == etree.tostring(real_etree)

    def test_get_saml_doc_different_location(self):
        a = SAMLAuthenticator()
        a.login_post_field = 'test'
        fake_data = {a.login_post_field: test_constants.b64encoded_response_xml}

        faked_etree = a._get_saml_doc_etree(fake_data)
        real_etree = etree.fromstring(test_constants.sample_response_xml)

        assert etree.tostring(faked_etree) == etree.tostring(real_etree)

    def test_with_failed_get(self):
        a = SAMLAuthenticator()
        fake_data = {}

        assert a._get_saml_doc_etree(fake_data) is None

    def test_bad_decode(self):
        a = SAMLAuthenticator()
        fake_data = {a.login_post_field: 'this is not base 64 encoded data'}

        assert a._get_saml_doc_etree(fake_data) is None

    @patch('samlauthenticator.samlauthenticator.b64decode')
    def test_not_valid_xml(self, mock_b64decode):
        a = SAMLAuthenticator()
        fake_data = {a.login_post_field: 'this string isn\'t important'}

        mock_b64decode.return_value = 'bad xml string'

        assert a._get_saml_doc_etree(fake_data) is None


class TestValidSamlResponse(unittest.TestCase):
    response_etree = etree.fromstring(test_constants.sample_response_xml)
    metadata_etree = etree.fromstring(test_constants.sample_metadata_xml)
    verified_signed_xml = XMLVerifier().verify(response_etree, x509_cert=test_constants.x509_cert).signed_xml

    @patch('samlauthenticator.samlauthenticator.datetime')
    def test_valid_saml_auth(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2019, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
        mock_datetime.strptime = datetime.strptime

        a = SAMLAuthenticator()

        signed_xml = a._verify_saml_signature(self.metadata_etree, self.response_etree)

        assert etree.tostring(signed_xml) == etree.tostring(self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)

        assert response_is_valid
        # Check the signed xml is the subset of the xml that is returned by signxml
        assert etree.tostring(signed_xml) == etree.tostring(self.verified_signed_xml)

    def test_tampered_saml_response(self):
        a = SAMLAuthenticator()
        tampered_etree = etree.fromstring(test_constants.tampered_sample_response_xml)

        bad_signed_xml = a._verify_saml_signature(self.metadata_etree, tampered_etree)

        assert bad_signed_xml is None

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, tampered_etree)

        assert not response_is_valid
        assert signed_xml is None

    def test_no_metadata_cert(self):
        a = SAMLAuthenticator()
        no_cert_metadata_etree = etree.fromstring(test_constants.sample_metadata_no_cert_xml)

        bad_signed_xml = a._verify_saml_signature(no_cert_metadata_etree, self.response_etree)

        assert bad_signed_xml is None

        response_is_valid, signed_xml = a._test_valid_saml_response(no_cert_metadata_etree, self.response_etree)

        assert not response_is_valid
        assert signed_xml is None

    def test_metadata_entity_no_match(self):
        a = SAMLAuthenticator()
        tampered_metadata_etree = etree.fromstring(test_constants.sample_metadata_tampered_entity)

        assert a._verify_saml_response_against_metadata(tampered_metadata_etree, self.verified_signed_xml) is False

        assert a._verify_saml_response_fields(tampered_metadata_etree, self.verified_signed_xml) is False

        response_is_valid, signed_xml = a._test_valid_saml_response(tampered_metadata_etree, self.response_etree)

        assert not response_is_valid
        assert etree.tostring(signed_xml) == etree.tostring(self.verified_signed_xml)

    def test_metadata_no_entity(self):
        a = SAMLAuthenticator()
        no_metadata_entity_etree = etree.fromstring(test_constants.sample_metadata_no_entity)

        assert a._verify_saml_response_against_metadata(no_metadata_entity_etree, self.verified_signed_xml) is False

        assert a._verify_saml_response_fields(no_metadata_entity_etree, self.verified_signed_xml) is False

        response_is_valid, signed_xml = a._test_valid_saml_response(no_metadata_entity_etree, self.response_etree)

        assert not response_is_valid
        assert etree.tostring(signed_xml) == etree.tostring(self.verified_signed_xml)

    def test_assertion_no_issuer(self):
        a = SAMLAuthenticator()

        tampered_etree = etree.fromstring(test_constants.tampered_assertion_no_issuer)

        assert not a._verify_saml_response_against_metadata(self.metadata_etree, tampered_etree)
        assert not a._verify_saml_response_fields(self.metadata_etree, tampered_etree)

    def test_signed_xml_bad_audience(self):
        a = SAMLAuthenticator()
        a.audience = '''bad_audience'''

        assert not a._verify_saml_response_against_configured_fields(self.verified_signed_xml)
        assert not a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert not response_is_valid
        # We will get the signed xml back, but the response is not valid, so it doesn't really matter
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    def test_signed_xml_no_audience(self):
        a = SAMLAuthenticator()
        a.audience = '''audience_should_exist'''

        tampered_etree = etree.fromstring(test_constants.tampered_assertion_no_audience)

        assert not a._verify_saml_response_against_configured_fields(tampered_etree)
        assert not a._verify_saml_response_fields(self.metadata_etree, tampered_etree)

    @patch('samlauthenticator.samlauthenticator.datetime')
    def test_signed_xml_good_audience(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2019, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
        mock_datetime.strptime = datetime.strptime

        a = SAMLAuthenticator()
        a.audience = '''{audience}'''

        assert a._verify_saml_response_against_configured_fields(self.verified_signed_xml)
        assert a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert response_is_valid
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    @patch('samlauthenticator.samlauthenticator.datetime')
    def test_signed_xml_good_recipient(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2019, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
        mock_datetime.strptime = datetime.strptime

        a = SAMLAuthenticator()
        a.recipient = '''{recipient}'''

        assert a._verify_saml_response_against_configured_fields(self.verified_signed_xml)
        assert a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert response_is_valid
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    def test_signed_xml_bad_recipient(self):
        a = SAMLAuthenticator()
        a.recipient = 'bad_recipient'

        assert not a._verify_saml_response_against_configured_fields(self.verified_signed_xml)
        assert not a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert not response_is_valid
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    def test_signed_xml_no_recipient(self):
        a = SAMLAuthenticator()
        a.recipient = 'unimportant_recipient'

        tampered_etree = etree.fromstring(test_constants.tampered_assertion_no_recipient)

        assert not a._verify_saml_response_against_configured_fields(tampered_etree)
        assert not a._verify_saml_response_fields(self.metadata_etree, tampered_etree)

    @patch('samlauthenticator.samlauthenticator.datetime')
    def test_now_before_allowed(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2018, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
        mock_datetime.strptime = datetime.strptime

        a = SAMLAuthenticator()

        assert not a._verify_physical_constraints(self.verified_signed_xml)
        assert not a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert not response_is_valid
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    @patch('samlauthenticator.samlauthenticator.datetime')
    def test_now_after_allowed(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2020, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
        mock_datetime.strptime = datetime.strptime

        a = SAMLAuthenticator()

        assert not a._verify_physical_constraints(self.verified_signed_xml)
        assert not a._verify_saml_response_fields(self.metadata_etree, self.verified_signed_xml)

        response_is_valid, signed_xml = a._test_valid_saml_response(self.metadata_etree, self.response_etree)
        assert not response_is_valid
        assert etree.tostring(self.verified_signed_xml) == etree.tostring(signed_xml)

    def test_no_not_on_or_after(self):
        a = SAMLAuthenticator()

        tampered_etree = etree.fromstring(test_constants.tampered_assertion_no_on_or_after)

        assert not a._verify_physical_constraints(tampered_etree)
        assert not a._verify_saml_response_fields(self.metadata_etree, tampered_etree)

    def test_no_not_before(self):
        a = SAMLAuthenticator()

        tampered_etree = etree.fromstring(test_constants.tampered_assertion_no_not_before)

        assert not a._verify_physical_constraints(tampered_etree)
        assert not a._verify_saml_response_fields(self.metadata_etree, tampered_etree)


class TestGetUsername(unittest.TestCase):
    response_etree = etree.fromstring(test_constants.sample_response_xml)
    verified_signed_xml = XMLVerifier().verify(response_etree, x509_cert=test_constants.x509_cert).signed_xml

    def test_get_username_from_saml_doc(self):
        a = SAMLAuthenticator()

        assert 'Bluedata' == a._get_username_from_saml_etree(self.verified_signed_xml)
        assert 'Bluedata' == a._get_username_from_saml_etree(self.response_etree)
        assert 'Bluedata' == a._get_username_from_saml_doc(self.verified_signed_xml, self.response_etree)

        a.xpath_username_location = 'string({})'.format(a.xpath_username_location)
        assert 'Bluedata' == a._get_username_from_saml_etree(self.verified_signed_xml)
        assert 'Bluedata' == a._get_username_from_saml_etree(self.response_etree)
        assert 'Bluedata' == a._get_username_from_saml_doc(self.verified_signed_xml, self.response_etree)

        a.xpath_username_location = 'substring-before({}, "data")'.format(a.xpath_username_location)
        assert 'Blue' == a._get_username_from_saml_etree(self.verified_signed_xml)
        assert 'Blue' == a._get_username_from_saml_etree(self.response_etree)
        assert 'Blue' == a._get_username_from_saml_doc(self.verified_signed_xml, self.response_etree)

    def test_get_username_no_nameid(self):
        tampered_assertion_etree = etree.fromstring(test_constants.tampered_assertion_no_nameid)
        tampered_response_etree  = etree.fromstring(test_constants.tampered_response_no_nameid)

        a = SAMLAuthenticator()

        assert a._get_username_from_saml_etree(tampered_assertion_etree) is None
        assert a._get_username_from_saml_etree(tampered_response_etree) is None
        assert a._get_username_from_saml_doc(tampered_assertion_etree, tampered_response_etree) is None
        assert 'Bluedata' == a._get_username_from_saml_doc(tampered_assertion_etree, self.response_etree)


class TestGetSingleRole(unittest.TestCase):
    response_etree = etree.fromstring(test_constants.sample_response_xml_with_one_role)
    verified_signed_xml = XMLVerifier().verify(response_etree, x509_cert=test_constants.x509_cert).signed_xml

    def test_get_roles_from_saml_doc(self):
        a = SAMLAuthenticator()
        a.xpath_role_location = '//saml:AttributeStatement/saml:Attribute[@Name="Roles"]/saml:AttributeValue/text()'

        assert ['Default'] == a._get_roles_from_saml_etree(self.verified_signed_xml)
        assert ['Default'] == a._get_roles_from_saml_etree(self.response_etree)
        assert ['Default'] == a._get_roles_from_saml_doc(self.verified_signed_xml, self.response_etree)

        a.xpath_role_location = '//saml:AttributeStatement/saml:Attribute[@Name="Bad_Role_Location"]/saml:AttributeValue/text()'

        assert a._get_roles_from_saml_etree(self.verified_signed_xml) == []
        assert a._get_roles_from_saml_etree(self.response_etree) == []
        assert a._get_roles_from_saml_doc(self.verified_signed_xml, self.response_etree) == []


class TestGetManyRoles(unittest.TestCase):
    response_etree = etree.fromstring(test_constants.sample_response_xml_with_many_roles)
    verified_signed_xml = XMLVerifier().verify(response_etree, x509_cert=test_constants.x509_cert).signed_xml
    sorted_roles = sorted(['Default', 'Role_1'])

    def test_get_roles_from_saml_doc(self):
        a = SAMLAuthenticator()
        a.xpath_role_location = '//saml:AttributeStatement/saml:Attribute[@Name="Roles"]/saml:AttributeValue/text()'

        assert self.sorted_roles == sorted(a._get_roles_from_saml_etree(self.verified_signed_xml))
        assert self.sorted_roles == sorted(a._get_roles_from_saml_etree(self.response_etree))
        assert self.sorted_roles == sorted(a._get_roles_from_saml_doc(self.verified_signed_xml, self.response_etree))


class TestRoleAccess(unittest.TestCase):

    def test_check_role(self):
        a = SAMLAuthenticator()
        a.allowed_roles = 'group1'

        assert a._check_role(['group1'])
        assert a._check_role(['group1', 'group2'])

    def test_check_roles(self):
        a = SAMLAuthenticator()
        a.allowed_roles='group1, group2, group3'

        assert a._check_role(['group2'])
        assert a._check_role(['group2', 'group3'])
        assert a._check_role(['group1', 'nogroup1'])

    def test_check_role_fails(self):
        a = SAMLAuthenticator()
        a.allowed_roles='group1,group2,group3'

        assert not a._check_role([])
        assert not a._check_role(['nogroup1'])
        assert not a._check_role(['nogroup1', 'nogroup2'])


class TestValidRolesConfig(unittest.TestCase):

    def test_no_xpath_no_roles_run_default(self):
        a = SAMLAuthenticator()
        a._valid_roles_in_assertion = unittest.mock.create_autospec(MagicMock(name='_valid_roles_in_assertion'))
        a.log.warning = MagicMock(name='warning')

        assert a._valid_config_and_roles(None, None)
        a._valid_roles_in_assertion.assert_not_called()
        a.log.warning.assert_not_called()

    def test_xpath_roles_call_methods_true_return(self):
        a = SAMLAuthenticator()
        a._valid_roles_in_assertion = MagicMock(name='_valid_roles_in_assertion', return_value=True)
        a.log.warning = MagicMock(name='warning')
        a.allowed_roles = 'group1'
        a.xpath_role_location = 'value'

        assert a._valid_config_and_roles(None, None) == True
        a._valid_roles_in_assertion.assert_called_once_with(None, None)
        a.log.warning.assert_not_called()

    def test_xpath_roles_call_methods_false_return(self):
        a = SAMLAuthenticator()
        a._valid_roles_in_assertion = MagicMock(name='_valid_roles_in_assertion', return_value=False)
        a.log.warning = MagicMock(name='warning')
        a.allowed_roles = 'group1'
        a.xpath_role_location = 'value'

        assert a._valid_config_and_roles(None, None) == False
        a._valid_roles_in_assertion.assert_called_once_with(None, None)
        a.log.warning.assert_not_called()

    def test_xpath_no_roles(self):
        a = SAMLAuthenticator()
        a.xpath_role_location = 'value'
        a._valid_roles_in_assertion = unittest.mock.create_autospec(MagicMock(name='_valid_roles_in_assertion'))
        a.log.warning = MagicMock(name='warning')

        assert a._valid_config_and_roles(None, None)
        a._valid_roles_in_assertion.assert_not_called()
        print(a.log.warning.call_args_list)
        a.log.warning.assert_called()
        a.log.warning.assert_any_call(a._const_warn_explain)
        a.log.warning.assert_any_call(a._const_warn_no_roles)

    def test_no_xpath_roles(self):
        a = SAMLAuthenticator()
        a.allowed_roles = 'value'
        a._valid_roles_in_assertion = unittest.mock.create_autospec(MagicMock(name='_valid_roles_in_assertion'))
        a.log.warning = MagicMock(name='warning')

        assert a._valid_config_and_roles(None, None)
        a._valid_roles_in_assertion.assert_not_called()
        print(a.log.warning.call_args_list)
        a.log.warning.assert_called()
        a.log.warning.assert_any_call(a._const_warn_explain)
        a.log.warning.assert_any_call(a._const_warn_no_role_xpath)


class TestCreateUser(unittest.TestCase):
    @patch('samlauthenticator.samlauthenticator.subprocess')
    @patch('samlauthenticator.samlauthenticator.pwd')
    def test_create_existing_user(self, mock_pwd, mock_subprocess):
        mock_pwd.getpwnam.return_value = True

        a = SAMLAuthenticator()

        assert a._optional_user_add('Bluedata')

        mock_pwd.getpwnam.assert_called_once_with('Bluedata')
        mock_subprocess.call.assert_not_called()

    @patch('samlauthenticator.samlauthenticator.subprocess')
    @patch('samlauthenticator.samlauthenticator.pwd')
    def test_create_not_existing_user(self, mock_pwd, mock_subprocess):
        mock_pwd.getpwnam.side_effect = KeyError('Bad username')
        mock_subprocess.call.return_value = 0

        a = SAMLAuthenticator()

        assert a._optional_user_add('Bluedata')

        mock_pwd.getpwnam.assert_called_once_with('Bluedata')
        mock_subprocess.call.assert_called_once_with(['useradd', 'Bluedata'])

    @patch('samlauthenticator.samlauthenticator.subprocess')
    @patch('samlauthenticator.samlauthenticator.pwd')
    def test_create_user_fails(self, mock_pwd, mock_subprocess):
        mock_pwd.getpwnam.side_effect = KeyError('Bad username')
        mock_subprocess.call.return_value = 1

        a = SAMLAuthenticator()

        assert not a._optional_user_add('Bluedata')

        mock_pwd.getpwnam.assert_called_once_with('Bluedata')
        mock_subprocess.call.assert_called_once_with(['useradd', 'Bluedata'])

    @patch('samlauthenticator.samlauthenticator.subprocess')
    @patch('samlauthenticator.samlauthenticator.pwd')
    def test_create_user_alternate_binary(self, mock_pwd, mock_subprocess):
        mock_pwd.getpwnam.side_effect = KeyError('Bad username')
        mock_subprocess.call.return_value = 0
        binary_value = 'test_binary'

        a = SAMLAuthenticator()
        a.create_system_user_binary = binary_value

        assert a._optional_user_add('Bluedata')

        mock_pwd.getpwnam.assert_called_once_with('Bluedata')
        mock_subprocess.call.assert_called_once_with([binary_value, 'Bluedata'])

    @patch('samlauthenticator.samlauthenticator.subprocess')
    @patch('samlauthenticator.samlauthenticator.pwd')
    def test_create_user_alternate_binary_existing_user(self, mock_pwd, mock_subprocess):
        mock_pwd.getpwnam.return_value = True
        binary_value = 'test_binary'

        a = SAMLAuthenticator()
        a.create_system_user_binary = binary_value

        assert a._optional_user_add('Bluedata')

        mock_pwd.getpwnam.assert_called_once_with('Bluedata')
        mock_subprocess.call.assert_not_called()

    def test_check_username_valid_username_no_black_lists(self):
        a = SAMLAuthenticator()
        a._optional_user_add = MagicMock()
        a._optional_user_add.return_value = True
        a.whitelist = {'bluedata'}

        assert a._check_username_and_add_user('bluedata')

        a._optional_user_add.assert_called_once_with('bluedata')

        a.whitelist = {'not_bluedata'}
        a._optional_user_add.reset_mock()

        assert not a._check_username_and_add_user('bluedata')

        a._optional_user_add.assert_not_called()

    def test_check_username_valid_username_no_white_lists(self):
        a = SAMLAuthenticator()
        a._optional_user_add = MagicMock()
        a._optional_user_add.return_value = True
        a.blacklist = {'bluedata'}

        assert not a._check_username_and_add_user('bluedata')

        a._optional_user_add.assert_not_called()

        a.blacklist = {'not_bluedata'}

        assert a._check_username_and_add_user('bluedata')

        a._optional_user_add.assert_called_once_with('bluedata')

    def test_check_username_invalid_username(self):
        a = SAMLAuthenticator()
        a._optional_user_add = MagicMock()

        assert not a._check_username_and_add_user('bluedata/')

        a._optional_user_add.assert_not_called()

    def test_create_system_users_option(self):
        a = SAMLAuthenticator()
        a.create_system_users = False
        a._optional_user_add = MagicMock()

        assert a._check_username_and_add_user('bluedata')

        a._optional_user_add.assert_not_called()


class TestAuthenticate(unittest.TestCase):
    def _confirm_tom(self, saml_data, mock_datetime, mock_pwd):
        mock_datetime.now.return_value = saml_data.datetime_stamp
        mock_datetime.strptime = datetime.strptime
        mock_pwd.getpwnam.return_value = True

        a = SAMLAuthenticator()
        a.metadata_content = saml_data.metadata_xml

        assert 'tom' == a._authenticate(None, {a.login_post_field: saml_data.b64encoded_response})
        mock_datetime.now.assert_called_once_with(timezone.utc)
        mock_pwd.getpwnam.assert_called_once_with('tom')

    def test_low_strength_cert_sha_1_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['low_strength_cert']['SHA-1']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_low_strength_cert_sha_256_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['low_strength_cert']['SHA-256']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_low_strength_cert_sha_384_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['low_strength_cert']['SHA-384']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_low_strength_cert_sha_512_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['low_strength_cert']['SHA-512']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_standard_strength_cert_sha_1_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['standard_strength_cert']['SHA-1']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_standard_strength_cert_sha_256_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['standard_strength_cert']['SHA-256']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_standard_strength_cert_sha_384_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['standard_strength_cert']['SHA-384']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_standard_strength_cert_sha_512_fingerprint(self):
        saml_data = test_constants.metadata_encoded_xml_dict['standard_strength_cert']['SHA-512']
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd:
            self._confirm_tom(saml_data, mock_datetime, mock_pwd)

    def test_bad_post_data(self):
        a = SAMLAuthenticator()
        # None because we can't get the response
        assert a._authenticate(None, {}) is None

    def test_bad_metadata_config(self):
        a = SAMLAuthenticator()
        # None because we can't get the metadata
        assert a._authenticate(None, {a.login_post_field: test_constants.b64encoded_response_xml}) is None

    def test_tampered_response(self):
        a = SAMLAuthenticator()
        a.metadata_content = test_constants.sample_metadata_xml
        assert a._authenticate(None, {a.login_post_field: test_constants.tampered_sample_response_encoded}) is None

    def test_no_allowed_roles(self):
        with patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime:
            mock_datetime.now.return_value = datetime(2020, 7, 1, 23, 0, 0, tzinfo=timezone.utc)
            mock_datetime.strptime = datetime.strptime
            a = SAMLAuthenticator()
            a.metadata_content = test_constants.sample_metadata_xml
            a.xpath_role_location = '//saml:AttributeStatement/saml:Attribute[@Name="Roles"]/saml:AttributeValue/text()'
            # The included XML should not have either of these roles.
            a.allowed_roles = 'allowed_role_1,allowed_role_2'
            assert a._authenticate(None, {a.login_post_field: test_constants.b64encoded_response_xml_with_roles}) is None
            mock_datetime.now.assert_called_once_with(timezone.utc)

    def test_add_user_fail(self):
        with patch('samlauthenticator.samlauthenticator.pwd') as mock_pwd, \
                patch('samlauthenticator.samlauthenticator.datetime') as mock_datetime, \
                patch('samlauthenticator.samlauthenticator.subprocess') as mock_subprocess:
            mock_pwd.getpwnam.side_effect = KeyError('No User')
            mock_datetime.now.return_value = datetime(2019, 4, 9, 21, 35, 0, tzinfo=timezone.utc)
            mock_datetime.strptime = datetime.strptime
            mock_subprocess.call.return_value = 1
            a = SAMLAuthenticator()
            a.metadata_content = test_constants.sample_metadata_xml
            assert a._authenticate(None, {a.login_post_field: test_constants.b64encoded_response_xml}) is None
            mock_pwd.getpwnam.assert_called_once_with('bluedata')
            mock_datetime.now.assert_called_once_with(timezone.utc)
            mock_subprocess.call.assert_called_once_with(['useradd', 'bluedata'])


class TestGetRedirect(unittest.TestCase):

    def test_get_valid_login_redirect(self):
        a = SAMLAuthenticator()
        a.metadata_content = test_constants.sample_metadata_xml

        mock_handler_self = MagicMock()

        a._get_redirect_from_metadata_and_redirect('md:SingleSignOnService', mock_handler_self)

        mock_handler_self.redirect.assert_called_once_with(
            'https://bluedata-test-before-deploy.onelogin.com/trust/saml2/http-redirect/sso/719630',
            permanent=False)

    def test_get_valid_logout_redirect(self):
        a = SAMLAuthenticator()
        a.metadata_content = test_constants.sample_metadata_xml

        mock_handler_self = MagicMock()

        a._get_redirect_from_metadata_and_redirect('md:SingleLogoutService', mock_handler_self)

        mock_handler_self.redirect.assert_called_once_with(
            'https://bluedata-test-before-deploy.onelogin.com/trust/saml2/http-redirect/slo/719630',
            permanent=False)

    def test_get_invalid_xml_element(self):
        a = SAMLAuthenticator()
        a.metadata_content = test_constants.sample_metadata_xml

        mock_handler_self = MagicMock()

        with self.assertRaises(IndexError):
            a._get_redirect_from_metadata_and_redirect('md:BadElement', mock_handler_self)

    def test_get_empty_metadata(self):
        a = SAMLAuthenticator()
        a.metadata_filepath = None
        a.metadata_content = None
        a.metadata_url = None

        mock_handler_self = MagicMock()

        with self.assertRaises(HTTPError):
            a._get_redirect_from_metadata_and_redirect('md:BadElement', mock_handler_self)


class TestMakeSPMetadata(unittest.TestCase):
    org_name_org_metadata = '''
    <Organization>
        <OrganizationName>org_name</OrganizationName>
        
        
    </Organization>
    '''
    org_display_name_org_metadata = '''
    <Organization>
        
        <OrganizationDisplayName>org_display_name</OrganizationDisplayName>
        
    </Organization>
    '''
    org_url_org_metadata = '''
    <Organization>
        
        
        <OrganizationURL>org_url</OrganizationURL>
    </Organization>
    '''
    full_sp_meta_default = '''<?xml version="1.0"?>
<EntityDescriptor
        entityID="https://localhost:8000"
        xmlns="urn:oasis:names:tc:SAML:2.0:metadata"
        xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
    <SPSSODescriptor
            AuthnRequestsSigned="false"
            protocolSupportEnumeration="urn:oasis:names:tc:SAML:2.0:protocol">
        <NameIDFormat>
            urn:oasis:names:tc:SAML:2.0:nameid-format:transient
        </NameIDFormat>
        <AssertionConsumerService
                Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                Location="https://localhost:8000/hub/login"/>
    </SPSSODescriptor>
    
</EntityDescriptor>'''
    full_sp_meta_nameid_format = '''<?xml version="1.0"?>
<EntityDescriptor
        entityID="https://localhost:8000"
        xmlns="urn:oasis:names:tc:SAML:2.0:metadata"
        xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
    <SPSSODescriptor
            AuthnRequestsSigned="false"
            protocolSupportEnumeration="urn:oasis:names:tc:SAML:2.0:protocol">
        <NameIDFormat>
            urn:oasis:names:tc:SAML:2.0:nameid-format:persistent
        </NameIDFormat>
        <AssertionConsumerService
                Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                Location="https://localhost:8000/hub/login"/>
    </SPSSODescriptor>
    
</EntityDescriptor>'''
    full_sp_meta_org_name = '''<?xml version="1.0"?>
<EntityDescriptor
        entityID="https://localhost:8000"
        xmlns="urn:oasis:names:tc:SAML:2.0:metadata"
        xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
    <SPSSODescriptor
            AuthnRequestsSigned="false"
            protocolSupportEnumeration="urn:oasis:names:tc:SAML:2.0:protocol">
        <NameIDFormat>
            urn:oasis:names:tc:SAML:2.0:nameid-format:transient
        </NameIDFormat>
        <AssertionConsumerService
                Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                Location="https://localhost:8000/hub/login"/>
    </SPSSODescriptor>
    
    <Organization>
        <OrganizationName>org_name</OrganizationName>
        
        
    </Organization>
    
</EntityDescriptor>'''

    def test_make_org_metadata_no_org_info(self):
        a = SAMLAuthenticator()
        a.organization_name = ''
        a.organization_display_name = ''
        a.organization_url = ''

        assert a._make_org_metadata() == ''

    def test_make_org_metadata_org_name(self):
        a = SAMLAuthenticator()
        a.organization_name = 'org_name'
        a.organization_display_name = ''
        a.organization_url = ''

        assert a._make_org_metadata() == self.org_name_org_metadata

    def test_make_org_metadata_org_display_name(self):
        a = SAMLAuthenticator()
        a.organization_name = ''
        a.organization_display_name = 'org_display_name'
        a.organization_url = ''

        assert a._make_org_metadata() == self.org_display_name_org_metadata

    def test_make_org_metadata_org_url(self):
        a = SAMLAuthenticator()
        a.organization_name = ''
        a.organization_display_name = ''
        a.organization_url = 'org_url'

        assert a._make_org_metadata() == self.org_url_org_metadata

    def test_make_full_metadata_default(self):
        a = SAMLAuthenticator()

        mock_handler_self = MagicMock()
        mock_handler_self.request.protocol = 'https'
        mock_handler_self.request.host = 'localhost:8000'

        assert a._make_sp_metadata(mock_handler_self) == self.full_sp_meta_default

    def test_make_full_metadata_org_name(self):
        a = SAMLAuthenticator()
        a.organization_name = 'org_name'

        mock_handler_self = MagicMock()
        mock_handler_self.request.protocol = 'https'
        mock_handler_self.request.host = 'localhost:8000'

        assert a._make_sp_metadata(mock_handler_self) == self.full_sp_meta_org_name

    def test_make_full_metadata_nameid_format(self):
        a = SAMLAuthenticator()
        a.nameid_format = 'urn:oasis:names:tc:SAML:2.0:nameid-format:persistent'

        mock_handler_self = MagicMock()
        mock_handler_self.request.protocol = 'https'
        mock_handler_self.request.host = 'localhost:8000'

        assert a._make_sp_metadata(mock_handler_self) == self.full_sp_meta_nameid_format


class TestGetHandlers(unittest.TestCase):
    expected_handler_paths = ['/login', '/hub/login', '/logout', '/hub/logout', '/metadata', '/hub/metadata']

    def test_get_handlers(self):
        a = SAMLAuthenticator()

        handler_list = a.get_handlers(None)

        for url_path, handler_class in handler_list:
            assert url_path in self.expected_handler_paths
            assert isinstance(handler_class, type)
            assert issubclass(handler_class, BaseHandler)
